# SPDX-License-Identifier: MIT
#
# Copyright (C) 2019 Philippe Proulx <pproulx@efficios.com>
#

# pyright: strict, reportTypeCommentUsage=false,  reportMissingTypeStubs=false

import os
import re
import enum
import time
import socket
import struct
import logging
import os.path
import argparse
import tempfile
from abc import ABC, abstractmethod
from typing import Dict, Union, Iterable, Optional, Sequence, overload

import tjson

# isort: off
from typing import Any, Callable  # noqa: F401

# isort: on


# An entry within the index of an LTTng data stream.
class _LttngDataStreamIndexEntry:
    def __init__(
        self,
        offset_bytes: int,
        total_size_bits: int,
        content_size_bits: int,
        timestamp_begin: int,
        timestamp_end: int,
        events_discarded: int,
        stream_class_id: int,
    ):
        self._offset_bytes = offset_bytes
        self._total_size_bits = total_size_bits
        self._content_size_bits = content_size_bits
        self._timestamp_begin = timestamp_begin
        self._timestamp_end = timestamp_end
        self._events_discarded = events_discarded
        self._stream_class_id = stream_class_id

    @property
    def offset_bytes(self):
        return self._offset_bytes

    @property
    def total_size_bits(self):
        return self._total_size_bits

    @property
    def total_size_bytes(self):
        return self._total_size_bits // 8

    @property
    def content_size_bits(self):
        return self._content_size_bits

    @property
    def content_size_bytes(self):
        return self._content_size_bits // 8

    @property
    def timestamp_begin(self):
        return self._timestamp_begin

    @property
    def timestamp_end(self):
        return self._timestamp_end

    @property
    def events_discarded(self):
        return self._events_discarded

    @property
    def stream_class_id(self):
        return self._stream_class_id


# An entry within the index of an LTTng data stream. While a stream beacon entry
# is conceptually unrelated to an index, it is sent as a reply to a
# LttngLiveViewerGetNextDataStreamIndexEntryCommand
class _LttngDataStreamBeaconIndexEntry:
    def __init__(self, stream_class_id: int, timestamp: int):
        self._stream_class_id = stream_class_id
        self._timestamp = timestamp

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def stream_class_id(self):
        return self._stream_class_id


_LttngIndexEntryT = Union[_LttngDataStreamIndexEntry, _LttngDataStreamBeaconIndexEntry]


class _LttngLiveViewerCommand:
    def __init__(self, version: int):
        self._version = version

    @property
    def version(self):
        return self._version


class _LttngLiveViewerConnectCommand(_LttngLiveViewerCommand):
    def __init__(self, version: int, viewer_session_id: int, major: int, minor: int):
        super().__init__(version)
        self._viewer_session_id = viewer_session_id
        self._major = major
        self._minor = minor

    @property
    def viewer_session_id(self):
        return self._viewer_session_id

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor


class _LttngLiveViewerReply:
    pass


class _LttngLiveViewerConnectReply(_LttngLiveViewerReply):
    def __init__(self, viewer_session_id: int, major: int, minor: int):
        self._viewer_session_id = viewer_session_id
        self._major = major
        self._minor = minor

    @property
    def viewer_session_id(self):
        return self._viewer_session_id

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor


class _LttngLiveViewerGetTracingSessionInfosCommand(_LttngLiveViewerCommand):
    pass


# Trace format of a tracing session.
class _LttngLiveTraceFormat(enum.Enum):
    CTF_V1_8 = "ctf-1.8"
    CTF_V2_0 = "ctf-2.0"


class _LttngLiveViewerTracingSessionInfo:
    def __init__(
        self,
        tracing_session_id: int,
        live_timer_freq: int,
        client_count: int,
        trace_format: _LttngLiveTraceFormat,
        stream_count: int,
        hostname: str,
        name: str,
    ):
        self._tracing_session_id = tracing_session_id
        self._live_timer_freq = live_timer_freq
        self._client_count = client_count
        self._trace_format = trace_format
        self._stream_count = stream_count
        self._hostname = hostname
        self._name = name

    @property
    def tracing_session_id(self):
        return self._tracing_session_id

    @property
    def live_timer_freq(self):
        return self._live_timer_freq

    @property
    def client_count(self):
        return self._client_count

    @property
    def trace_format(self):
        return self._trace_format

    @property
    def stream_count(self):
        return self._stream_count

    @property
    def hostname(self):
        return self._hostname

    @property
    def name(self):
        return self._name


class _LttngLiveViewerGetTracingSessionInfosReply(_LttngLiveViewerReply):
    def __init__(
        self, tracing_session_infos: Sequence[_LttngLiveViewerTracingSessionInfo]
    ):
        self._tracing_session_infos = tracing_session_infos

    @property
    def tracing_session_infos(self):
        return self._tracing_session_infos


class _LttngLiveViewerAttachToTracingSessionCommand(_LttngLiveViewerCommand):
    class SeekType:
        BEGINNING = 1
        LAST = 2

    def __init__(
        self, version: int, tracing_session_id: int, offset: int, seek_type: int
    ):
        super().__init__(version)
        self._tracing_session_id = tracing_session_id
        self._offset = offset
        self._seek_type = seek_type

    @property
    def tracing_session_id(self):
        return self._tracing_session_id

    @property
    def offset(self):
        return self._offset

    @property
    def seek_type(self):
        return self._seek_type


class _LttngLiveViewerStreamInfo:
    def __init__(
        self, id: int, trace_id: int, is_metadata: bool, path: str, channel_name: str
    ):
        self._id = id
        self._trace_id = trace_id
        self._is_metadata = is_metadata
        self._path = path
        self._channel_name = channel_name

    @property
    def id(self):
        return self._id

    @property
    def trace_id(self):
        return self._trace_id

    @property
    def is_metadata(self):
        return self._is_metadata

    @property
    def path(self):
        return self._path

    @property
    def channel_name(self):
        return self._channel_name


class _LttngLiveViewerAttachToTracingSessionReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        ALREADY = 2
        UNKNOWN = 3
        NOT_LIVE = 4
        SEEK_ERROR = 5
        NO_SESSION = 6

    def __init__(self, status: int, stream_infos: Sequence[_LttngLiveViewerStreamInfo]):
        self._status = status
        self._stream_infos = stream_infos

    @property
    def status(self):
        return self._status

    @property
    def stream_infos(self):
        return self._stream_infos


class _LttngLiveViewerGetNextDataStreamIndexEntryCommand(_LttngLiveViewerCommand):
    def __init__(self, version: int, stream_id: int):
        super().__init__(version)
        self._stream_id = stream_id

    @property
    def stream_id(self):
        return self._stream_id


class _LttngLiveViewerGetNextDataStreamIndexEntryReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        RETRY = 2
        HUP = 3
        ERROR = 4
        INACTIVE = 5
        EOF = 6

    def __init__(
        self,
        status: int,
        index_entry: _LttngIndexEntryT,
        has_new_metadata: bool,
        has_new_data_stream: bool,
    ):
        self._status = status
        self._index_entry = index_entry
        self._has_new_metadata = has_new_metadata
        self._has_new_data_stream = has_new_data_stream

    @property
    def status(self):
        return self._status

    @property
    def index_entry(self):
        return self._index_entry

    @property
    def has_new_metadata(self):
        return self._has_new_metadata

    @property
    def has_new_data_stream(self):
        return self._has_new_data_stream


class _LttngLiveViewerGetDataStreamPacketDataCommand(_LttngLiveViewerCommand):
    def __init__(self, version: int, stream_id: int, offset: int, req_length: int):
        super().__init__(version)
        self._stream_id = stream_id
        self._offset = offset
        self._req_length = req_length

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def offset(self):
        return self._offset

    @property
    def req_length(self):
        return self._req_length


class _LttngLiveViewerGetDataStreamPacketDataReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        RETRY = 2
        ERROR = 3
        EOF = 4

    def __init__(
        self,
        status: int,
        data: bytes,
        has_new_metadata: bool,
        has_new_data_stream: bool,
    ):
        self._status = status
        self._data = data
        self._has_new_metadata = has_new_metadata
        self._has_new_data_stream = has_new_data_stream

    @property
    def status(self):
        return self._status

    @property
    def data(self):
        return self._data

    @property
    def has_new_metadata(self):
        return self._has_new_metadata

    @property
    def has_new_data_stream(self):
        return self._has_new_data_stream


class _LttngLiveViewerGetMetadataStreamDataCommand(_LttngLiveViewerCommand):
    def __init__(self, version: int, stream_id: int):
        super().__init__(version)
        self._stream_id = stream_id

    @property
    def stream_id(self):
        return self._stream_id


class _LttngLiveViewerGetMetadataStreamDataContentReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        NO_NEW = 2
        ERROR = 3

    def __init__(self, status: int, data: bytes):
        self._status = status
        self._data = data

    @property
    def status(self):
        return self._status

    @property
    def data(self):
        return self._data


class _LttngLiveViewerGetNewStreamInfosCommand(_LttngLiveViewerCommand):
    def __init__(self, version: int, tracing_session_id: int):
        super().__init__(version)
        self._tracing_session_id = tracing_session_id

    @property
    def tracing_session_id(self):
        return self._tracing_session_id


class _LttngLiveViewerGetNewStreamInfosReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        NO_NEW = 2
        ERROR = 3
        HUP = 4

    def __init__(self, status: int, stream_infos: Sequence[_LttngLiveViewerStreamInfo]):
        self._status = status
        self._stream_infos = stream_infos

    @property
    def status(self):
        return self._status

    @property
    def stream_infos(self):
        return self._stream_infos


class _LttngLiveViewerCreateViewerSessionCommand(_LttngLiveViewerCommand):
    pass


class _LttngLiveViewerCreateViewerSessionReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        ERROR = 2

    def __init__(self, status: int):
        self._status = status

    @property
    def status(self):
        return self._status


class _LttngLiveViewerDetachFromTracingSessionCommand(_LttngLiveViewerCommand):
    def __init__(self, version: int, tracing_session_id: int):
        super().__init__(version)
        self._tracing_session_id = tracing_session_id

    @property
    def tracing_session_id(self):
        return self._tracing_session_id


class _LttngLiveViewerDetachFromTracingSessionReply(_LttngLiveViewerReply):
    class Status:
        OK = 1
        UNKNOWN = 2
        ERROR = 3

    def __init__(self, status: int):
        self._status = status

    @property
    def status(self):
        return self._status


# An LTTng live protocol codec can convert bytes to command objects and
# reply objects to bytes.
class _LttngLiveViewerProtocolCodec:
    _COMMAND_HEADER_STRUCT_FMT = "QII"
    _COMMAND_HEADER_SIZE_BYTES = struct.calcsize(_COMMAND_HEADER_STRUCT_FMT)

    def __init__(self):
        self._server_minor_version = None  # type: Optional[int]

    def _unpack(self, fmt: str, data: bytes, offset: int = 0):
        fmt = "!" + fmt
        return struct.unpack_from(fmt, data, offset)

    def _unpack_payload(self, fmt: str, data: bytes):
        return self._unpack(
            fmt, data, _LttngLiveViewerProtocolCodec._COMMAND_HEADER_SIZE_BYTES
        )

    @property
    def server_minor_version(self):
        return self._server_minor_version

    @server_minor_version.setter
    def server_minor_version(self, server_minor_version: int):
        self._server_minor_version = server_minor_version

    def decode(self, data: bytes):
        if len(data) < self._COMMAND_HEADER_SIZE_BYTES:
            # Not enough data to read the command header
            return

        payload_size, cmd_type, version = self._unpack(
            self._COMMAND_HEADER_STRUCT_FMT, data
        )
        logging.info(
            "Decoded command header: payload-size={}, cmd-type={}, version={}".format(
                payload_size, cmd_type, version
            )
        )

        if len(data) < self._COMMAND_HEADER_SIZE_BYTES + payload_size:
            # Not enough data to read the whole command
            return

        if cmd_type == 1:
            viewer_session_id, major, minor, _ = self._unpack_payload("QIII", data)
            return _LttngLiveViewerConnectCommand(
                version, viewer_session_id, major, minor
            )
        elif cmd_type == 2:
            return _LttngLiveViewerGetTracingSessionInfosCommand(version)
        elif cmd_type == 3:
            tracing_session_id, offset, seek_type = self._unpack_payload("QQI", data)
            return _LttngLiveViewerAttachToTracingSessionCommand(
                version, tracing_session_id, offset, seek_type
            )
        elif cmd_type == 4:
            (stream_id,) = self._unpack_payload("Q", data)
            return _LttngLiveViewerGetNextDataStreamIndexEntryCommand(
                version, stream_id
            )
        elif cmd_type == 5:
            stream_id, offset, req_length = self._unpack_payload("QQI", data)
            return _LttngLiveViewerGetDataStreamPacketDataCommand(
                version, stream_id, offset, req_length
            )
        elif cmd_type == 6:
            (stream_id,) = self._unpack_payload("Q", data)
            return _LttngLiveViewerGetMetadataStreamDataCommand(version, stream_id)
        elif cmd_type == 7:
            (tracing_session_id,) = self._unpack_payload("Q", data)
            return _LttngLiveViewerGetNewStreamInfosCommand(version, tracing_session_id)
        elif cmd_type == 8:
            return _LttngLiveViewerCreateViewerSessionCommand(version)
        elif cmd_type == 9:
            (tracing_session_id,) = self._unpack_payload("Q", data)
            return _LttngLiveViewerDetachFromTracingSessionCommand(
                version, tracing_session_id
            )
        else:
            raise RuntimeError("Unknown command type {}".format(cmd_type))

    def _pack(self, fmt: str, *args: Any):
        # Force network byte order
        return struct.pack("!" + fmt, *args)

    def _encode_zero_padded_str(self, string: str, length: int):
        data = string.encode()
        return data.ljust(length, b"\x00")

    def _encode_stream_info(self, info: _LttngLiveViewerStreamInfo):
        data = self._pack("QQI", info.id, info.trace_id, int(info.is_metadata))
        data += self._encode_zero_padded_str(info.path, 4096)
        data += self._encode_zero_padded_str(info.channel_name, 255)
        return data

    def _get_has_new_stuff_flags(
        self, has_new_metadata: bool, has_new_data_streams: bool
    ):
        flags = 0

        if has_new_metadata:
            flags |= 1

        if has_new_data_streams:
            flags |= 2

        return flags

    def encode(
        self,
        reply: _LttngLiveViewerReply,
    ) -> bytes:
        if type(reply) is _LttngLiveViewerConnectReply:
            data = self._pack(
                "QIII", reply.viewer_session_id, reply.major, reply.minor, 2
            )
        elif type(reply) is _LttngLiveViewerGetTracingSessionInfosReply:
            data = self._pack("I", len(reply.tracing_session_infos))

            for info in reply.tracing_session_infos:
                # Common part
                data += self._pack(
                    "QIII",
                    info.tracing_session_id,
                    info.live_timer_freq,
                    info.client_count,
                    info.stream_count,
                )

                # Version-specific part
                assert self._server_minor_version is not None

                if self._server_minor_version < 15:
                    # v2.4 protocol: fixed-length strings
                    data += self._encode_zero_padded_str(info.hostname, 64)
                    data += self._encode_zero_padded_str(info.name, 255)
                else:
                    # v2.15 protocol: variable-length strings + trace format
                    data += self._pack(
                        "III",
                        len(info.hostname),
                        len(info.name),
                        0 if info.trace_format == _LttngLiveTraceFormat.CTF_V1_8 else 1,
                    )
                    data += info.hostname.encode()
                    data += info.name.encode()
        elif type(reply) is _LttngLiveViewerAttachToTracingSessionReply:
            data = self._pack("II", reply.status, len(reply.stream_infos))

            for info in reply.stream_infos:
                data += self._encode_stream_info(info)
        elif type(reply) is _LttngLiveViewerGetNextDataStreamIndexEntryReply:
            index_format = "QQQQQQQII"
            entry = reply.index_entry
            flags = self._get_has_new_stuff_flags(
                reply.has_new_metadata, reply.has_new_data_stream
            )

            if isinstance(entry, _LttngDataStreamIndexEntry):
                data = self._pack(
                    index_format,
                    entry.offset_bytes,
                    entry.total_size_bits,
                    entry.content_size_bits,
                    entry.timestamp_begin,
                    entry.timestamp_end,
                    entry.events_discarded,
                    entry.stream_class_id,
                    reply.status,
                    flags,
                )
            else:
                data = self._pack(
                    index_format,
                    0,
                    0,
                    0,
                    0,
                    entry.timestamp,
                    0,
                    entry.stream_class_id,
                    reply.status,
                    flags,
                )
        elif type(reply) is _LttngLiveViewerGetDataStreamPacketDataReply:
            flags = self._get_has_new_stuff_flags(
                reply.has_new_metadata, reply.has_new_data_stream
            )
            data = self._pack("III", reply.status, len(reply.data), flags)
            data += reply.data
        elif type(reply) is _LttngLiveViewerGetMetadataStreamDataContentReply:
            data = self._pack("QI", len(reply.data), reply.status)
            data += reply.data
        elif type(reply) is _LttngLiveViewerGetNewStreamInfosReply:
            data = self._pack("II", reply.status, len(reply.stream_infos))

            for info in reply.stream_infos:
                data += self._encode_stream_info(info)
        elif type(reply) is _LttngLiveViewerCreateViewerSessionReply:
            data = self._pack("I", reply.status)
        elif type(reply) is _LttngLiveViewerDetachFromTracingSessionReply:
            data = self._pack("I", reply.status)
        else:
            raise ValueError(
                "Unknown reply object with class `{}`".format(reply.__class__.__name__)
            )

        return data


def _get_entry_timestamp_begin(
    entry: _LttngIndexEntryT,
):
    if isinstance(entry, _LttngDataStreamBeaconIndexEntry):
        return entry.timestamp
    else:
        return entry.timestamp_begin


# The index of an LTTng data stream, a sequence of index entries.
class _LttngDataStreamIndex(Sequence[_LttngIndexEntryT]):
    def __init__(self, path: str, beacons: Optional[tjson.ArrayVal]):
        self._path = path
        self._build()

        if beacons:
            stream_class_id = self._entries[0].stream_class_id

            beacons_list = []  # type: list[_LttngDataStreamBeaconIndexEntry]
            for ts in beacons.iter(tjson.IntVal):
                beacons_list.append(
                    _LttngDataStreamBeaconIndexEntry(stream_class_id, ts.val)
                )

            self._add_beacons(beacons_list)

        logging.info(
            'Built data stream index entries: path="{}", count={}'.format(
                path, len(self._entries)
            )
        )

    def _build(self):
        self._entries = []  # type: list[_LttngIndexEntryT]

        with open(self._path, "rb") as f:
            # Read header first
            fmt = ">IIII"
            size = struct.calcsize(fmt)
            data = f.read(size)
            assert len(data) == size
            magic, _, _, index_entry_length = struct.unpack(fmt, data)
            assert magic == 0xC1F1DCC1

            # Read index entries
            fmt = ">QQQQQQQ"
            size = struct.calcsize(fmt)

            while True:
                logging.debug(
                    'Decoding data stream index entry: path="{}", offset={}'.format(
                        self._path, f.tell()
                    )
                )
                data = f.read(size)

                if not data:
                    # Done
                    break

                assert len(data) == size
                (
                    offset_bytes,
                    total_size_bits,
                    content_size_bits,
                    timestamp_begin,
                    timestamp_end,
                    events_discarded,
                    stream_class_id,
                ) = struct.unpack(fmt, data)

                self._entries.append(
                    _LttngDataStreamIndexEntry(
                        offset_bytes,
                        total_size_bits,
                        content_size_bits,
                        timestamp_begin,
                        timestamp_end,
                        events_discarded,
                        stream_class_id,
                    )
                )

                # Skip anything else before the next entry
                f.seek(index_entry_length - size, os.SEEK_CUR)

    def _add_beacons(self, beacons: Iterable[_LttngDataStreamBeaconIndexEntry]):
        # Assumes entries[n + 1].timestamp_end >= entries[n].timestamp_begin
        def sort_key(
            entry: Union[_LttngDataStreamIndexEntry, _LttngDataStreamBeaconIndexEntry],
        ) -> int:
            if isinstance(entry, _LttngDataStreamBeaconIndexEntry):
                return entry.timestamp
            else:
                return entry.timestamp_end

        self._entries += beacons
        self._entries.sort(key=sort_key)

    @overload
    def __getitem__(self, index: int) -> _LttngIndexEntryT:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[_LttngIndexEntryT]:  # noqa: F811
        ...

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[_LttngIndexEntryT, Sequence[_LttngIndexEntryT],]:  # noqa: F811
        return self._entries[index]

    def __len__(self):
        return len(self._entries)

    @property
    def path(self):
        return self._path


# Any LTTng stream (metadata or data).
class _LttngStream(ABC):
    @abstractmethod
    def __init__(self, creation_timestamp: int):
        self._creation_timestamp = creation_timestamp

    @property
    def creation_timestamp(self):
        return self._creation_timestamp


# An LTTng data stream.
class _LttngDataStream(_LttngStream):
    def __init__(
        self, path: str, beacons_json: Optional[tjson.ArrayVal], creation_timestamp: int
    ):
        super().__init__(creation_timestamp)
        self._path = path
        filename = os.path.basename(path)
        match = re.match(r"(.*)_\d+", filename)
        if not match:
            raise RuntimeError(
                "Unexpected data stream file name pattern: {}".format(filename)
            )

        self._channel_name = match.group(1)
        trace_dir = os.path.dirname(path)
        index_path = os.path.join(trace_dir, "index", filename + ".idx")
        self._index = _LttngDataStreamIndex(index_path, beacons_json)
        assert os.path.isfile(path)
        self._file = open(path, "rb")
        logging.info(
            'Built data stream: path="{}", channel-name="{}"'.format(
                path, self._channel_name
            )
        )

    @property
    def path(self):
        return self._path

    @property
    def channel_name(self):
        return self._channel_name

    @property
    def index(self):
        return self._index

    def get_data(self, offset_bytes: int, len_bytes: int):
        self._file.seek(offset_bytes)
        return self._file.read(len_bytes)


class _LttngMetadataStreamSection:
    def __init__(self, timestamp: int, data: bytes):
        self._timestamp = timestamp
        self._data = data
        logging.info(
            "Built metadata stream section: ts={}, data-len={}".format(
                self._timestamp, len(self._data)
            )
        )

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def data(self):
        return self._data


# An LTTng metadata stream.
class _LttngMetadataStream(_LttngStream):
    def __init__(
        self,
        metadata_file_path: str,
        config_sections: Sequence[_LttngMetadataStreamSection],
        creation_timestamp: int,
    ):
        super().__init__(creation_timestamp)
        self._path = metadata_file_path
        self._sections = config_sections
        logging.info(
            "Built metadata stream: path={}, section-len={}".format(
                self._path, len(self._sections)
            )
        )

    @property
    def path(self):
        return self._path

    @property
    def sections(self):
        return self._sections

    @property
    def trace_format(self):
        with open(self._path, "rb") as f:
            return (
                _LttngLiveTraceFormat.CTF_V2_0
                if f.read(1)[0] == 30
                else _LttngLiveTraceFormat.CTF_V1_8
            )


class LttngMetadataConfigSection:
    def __init__(self, line: int, timestamp: int, is_empty: bool):
        self._line = line
        self._timestamp = timestamp
        self._is_empty = is_empty

    @property
    def line(self):
        return self._line

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def is_empty(self):
        return self._is_empty


def _parse_metadata_sections_config(metadata_sections_json: tjson.ArrayVal):
    metadata_sections = []  # type: list[LttngMetadataConfigSection]
    append_empty_section = False
    last_timestamp = 0
    last_line = 0

    for section in metadata_sections_json:
        if isinstance(section, tjson.StrVal):
            if section.val == "empty":
                # Found an empty section marker. Actually append the
                # section at the timestamp of the next concrete section.
                append_empty_section = True
            else:
                raise ValueError("Invalid string value at {}.".format(section.path))
        elif isinstance(section, tjson.ObjVal):
            line = section.at("line", tjson.IntVal).val
            ts = section.at("timestamp", tjson.IntVal).val

            # Sections' timestamps and lines must both be increasing.
            assert ts > last_timestamp
            last_timestamp = ts

            assert line > last_line
            last_line = line

            if append_empty_section:
                metadata_sections.append(LttngMetadataConfigSection(line, ts, True))
                append_empty_section = False

            metadata_sections.append(LttngMetadataConfigSection(line, ts, False))
        else:
            raise TypeError(
                "`{}`: expecting a string or object value".format(section.path)
            )

    return metadata_sections


def _split_metadata_sections(
    metadata_file_path: str, metadata_sections_json: tjson.ArrayVal
):
    metadata_sections = _parse_metadata_sections_config(metadata_sections_json)

    sections = []  # type: list[_LttngMetadataStreamSection]
    with open(metadata_file_path, "r") as metadata_file:
        metadata_lines = [line for line in metadata_file]

    metadata_section_idx = 0
    curr_metadata_section = bytearray()

    for idx, line_content in enumerate(metadata_lines):
        # Add one to the index to convert from the zero-indexing of the
        # enumerate() function to the one-indexing used by humans when
        # viewing a text file.
        curr_line_number = idx + 1

        # If there are no more sections, simply append the line.
        if metadata_section_idx + 1 >= len(metadata_sections):
            curr_metadata_section += bytearray(line_content, "utf8")
            continue

        next_section_line_number = metadata_sections[metadata_section_idx + 1].line

        # If the next section begins at the current line, create a
        # section with the metadata we gathered so far.
        if curr_line_number >= next_section_line_number:
            # Flushing the metadata of the current section.
            sections.append(
                _LttngMetadataStreamSection(
                    metadata_sections[metadata_section_idx].timestamp,
                    bytes(curr_metadata_section),
                )
            )

            # Move to the next section.
            metadata_section_idx += 1

            # Clear old content and append current line for the next section.
            curr_metadata_section.clear()
            curr_metadata_section += bytearray(line_content, "utf8")

            # Append any empty sections.
            while metadata_sections[metadata_section_idx].is_empty:
                sections.append(
                    _LttngMetadataStreamSection(
                        metadata_sections[metadata_section_idx].timestamp, bytes()
                    )
                )
                metadata_section_idx += 1
        else:
            # Append line_content to the current metadata section.
            curr_metadata_section += bytearray(line_content, "utf8")

    # We iterated over all the lines of the metadata file. Close the current section.
    sections.append(
        _LttngMetadataStreamSection(
            metadata_sections[metadata_section_idx].timestamp,
            bytes(curr_metadata_section),
        )
    )

    return sections


_StreamBeaconsT = Dict[str, Iterable[int]]


# An LTTng trace, a sequence of LTTng data streams.
class LttngTrace(Sequence[_LttngDataStream]):
    def __init__(
        self,
        trace_dir: str,
        metadata_sections_json: Optional[tjson.ArrayVal],
        beacons_json: Optional[tjson.ObjVal],
        creation_timestamp: int,
    ):
        self._path = trace_dir
        self._creation_timestamp = creation_timestamp
        self._create_metadata_stream(trace_dir, metadata_sections_json)
        self._create_data_streams(trace_dir, beacons_json)
        logging.info('Built trace: path="{}"'.format(trace_dir))

    def _create_data_streams(
        self, trace_dir: str, beacons_json: Optional[tjson.ObjVal]
    ):
        data_stream_paths = []  # type: list[str]

        for filename in os.listdir(trace_dir):
            path = os.path.join(trace_dir, filename)

            if not os.path.isfile(path):
                continue

            if filename.startswith("."):
                continue

            if filename == "metadata":
                continue

            data_stream_paths.append(path)

        data_stream_paths.sort()
        self._data_streams = []  # type: list[_LttngDataStream]

        for data_stream_path in data_stream_paths:
            stream_name = os.path.basename(data_stream_path)
            this_beacons_json = None
            if beacons_json is not None and stream_name in beacons_json:
                this_beacons_json = beacons_json.at(stream_name, tjson.ArrayVal)

            self._data_streams.append(
                _LttngDataStream(
                    data_stream_path, this_beacons_json, self._creation_timestamp
                )
            )

    def _create_metadata_stream(
        self, trace_dir: str, metadata_sections_json: Optional[tjson.ArrayVal]
    ):
        metadata_path = os.path.join(trace_dir, "metadata")
        metadata_sections = []  # type: list[_LttngMetadataStreamSection]

        if metadata_sections_json is None:
            with open(metadata_path, "rb") as metadata_file:
                metadata_sections.append(
                    _LttngMetadataStreamSection(0, metadata_file.read())
                )
        else:
            metadata_sections = _split_metadata_sections(
                metadata_path, metadata_sections_json
            )

        self._metadata_stream = _LttngMetadataStream(
            metadata_path, metadata_sections, self.creation_timestamp
        )

    @property
    def path(self):
        return self._path

    @property
    def metadata_stream(self):
        return self._metadata_stream

    @property
    def creation_timestamp(self):
        return self._creation_timestamp

    @overload
    def __getitem__(self, index: int) -> _LttngDataStream:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[_LttngDataStream]:  # noqa: F811
        ...

    def __getitem__(  # noqa: F811
        self, index: Union[int, slice]
    ) -> Union[_LttngDataStream, Sequence[_LttngDataStream]]:
        return self._data_streams[index]

    def __len__(self):
        return len(self._data_streams)


# Stream (metadata or data) state specific to the LTTng live protocol.
class _LttngLiveViewerSessionStreamState:
    @abstractmethod
    def __init__(self):
        # A stream is considered "announced" when it has been returned
        # to the LTTng live client in response to a "get new stream
        # infos" (`_LttngLiveViewerGetNewStreamInfosCommand`) command.
        self._announced = False  # type: bool
        pass

    @property
    def is_announced(self):
        return self._announced

    def mark_as_announced(self):
        self._announced = True

    @property
    @abstractmethod
    def stream(self) -> _LttngStream:
        pass


# The state of a single data stream.
class _LttngLiveViewerSessionDataStreamState(_LttngLiveViewerSessionStreamState):
    def __init__(
        self,
        ts_state: "_LttngLiveViewerSessionTracingSessionState",
        info: _LttngLiveViewerStreamInfo,
        data_stream: _LttngDataStream,
        metadata_stream_id: int,
    ):
        super().__init__()
        self._ts_state = ts_state
        self._info = info
        self._data_stream = data_stream
        self._metadata_stream_id = metadata_stream_id
        self._cur_index_entry_index = 0
        fmt = 'Built data stream state: id={}, ts-id={}, ts-name="{}", path="{}"'
        logging.info(
            fmt.format(
                info.id,
                ts_state.tracing_session_descriptor.info.tracing_session_id,
                ts_state.tracing_session_descriptor.info.name,
                data_stream.path,
            )
        )

    @property
    def tracing_session_state(self):
        return self._ts_state

    @property
    def info(self):
        return self._info

    @property
    def stream(self):
        return self._data_stream

    @property
    def cur_index_entry(self):
        if self._cur_index_entry_index == len(self._data_stream.index):
            return

        return self._data_stream.index[self._cur_index_entry_index]

    @property
    def metadata_stream_id(self):
        return self._metadata_stream_id

    def goto_next_index_entry(self):
        self._cur_index_entry_index += 1


# The state of a single metadata stream.
class _LttngLiveViewerSessionMetadataStreamState(_LttngLiveViewerSessionStreamState):
    def __init__(
        self,
        ts_state: "_LttngLiveViewerSessionTracingSessionState",
        info: _LttngLiveViewerStreamInfo,
        metadata_stream: _LttngMetadataStream,
    ):
        super().__init__()
        self._ts_state = ts_state
        self._info = info
        self._metadata_stream = metadata_stream
        self._cur_metadata_stream_section_index = 0
        if len(metadata_stream.sections) > 1:
            self._next_metadata_stream_section_timestamp = metadata_stream.sections[
                1
            ].timestamp
        else:
            self._next_metadata_stream_section_timestamp = None

        self._all_data_is_sent = False
        fmt = 'Built metadata stream state: id={}, ts-id={}, ts-name="{}", path="{}"'
        logging.info(
            fmt.format(
                info.id,
                ts_state.tracing_session_descriptor.info.tracing_session_id,
                ts_state.tracing_session_descriptor.info.name,
                metadata_stream.path,
            )
        )

    @property
    def info(self):
        return self._info

    @property
    def stream(self):
        return self._metadata_stream

    @property
    def all_data_is_sent(self):
        return self._all_data_is_sent

    @all_data_is_sent.setter
    def all_data_is_sent(self, value: bool):
        self._all_data_is_sent = value

    @property
    def cur_section(self):
        fmt = "Get current metadata section: section-idx={}"
        logging.info(fmt.format(self._cur_metadata_stream_section_index))
        if self._cur_metadata_stream_section_index == len(
            self._metadata_stream.sections
        ):
            return

        return self._metadata_stream.sections[self._cur_metadata_stream_section_index]

    def goto_next_section(self):
        self._cur_metadata_stream_section_index += 1
        if self.cur_section:
            self._next_metadata_stream_section_timestamp = self.cur_section.timestamp
        else:
            self._next_metadata_stream_section_timestamp = None

    @property
    def next_section_timestamp(self):
        return self._next_metadata_stream_section_timestamp


# A tracing session descriptor.
#
# In the constructor, `traces` is a list of LTTng traces (`LttngTrace`
# objects).
class LttngTracingSessionDescriptor:
    def __init__(
        self,
        name: str,
        tracing_session_id: int,
        hostname: str,
        live_timer_freq: int,
        client_count: int,
        trace_format: _LttngLiveTraceFormat,
        traces: Iterable[LttngTrace],
    ):
        for trace in traces:
            if name not in trace.path:
                fmt = "Tracing session name must be part of every trace path (`{}` not found in `{}`)"
                raise ValueError(fmt.format(name, trace.path))

        self._traces = traces
        stream_count = sum([len(t) + 1 for t in traces])
        self._info = _LttngLiveViewerTracingSessionInfo(
            tracing_session_id,
            live_timer_freq,
            client_count,
            trace_format,
            stream_count,
            hostname,
            name,
        )

    @property
    def traces(self):
        return self._traces

    @property
    def info(self):
        return self._info


# The state of a tracing session.
class _LttngLiveViewerSessionTracingSessionState:
    def __init__(self, tc_descr: LttngTracingSessionDescriptor, base_stream_id: int):
        self._tc_descr = tc_descr
        self._client_visible_stream_infos = []  # type: list[_LttngLiveViewerStreamInfo]
        self._ds_states = {}  # type: dict[int, _LttngLiveViewerSessionDataStreamState]
        self._ms_states = (
            {}
        )  # type: dict[int, _LttngLiveViewerSessionMetadataStreamState]
        self._last_delivered_index_timestamp = 0
        self._last_allocated_stream_id = base_stream_id

        for trace in tc_descr.traces:
            trace_stream_infos = []  # type: list[_LttngLiveViewerStreamInfo]
            trace_id = self._last_allocated_stream_id * 1000

            # Metadata stream -> stream info and metadata stream state
            info = _LttngLiveViewerStreamInfo(
                self._last_allocated_stream_id,
                trace_id,
                True,
                trace.metadata_stream.path,
                "metadata",
            )

            trace_stream_infos.append(info)
            self._ms_states[
                self._last_allocated_stream_id
            ] = _LttngLiveViewerSessionMetadataStreamState(
                self, info, trace.metadata_stream
            )
            metadata_stream_id = self._last_allocated_stream_id
            self._last_allocated_stream_id += 1

            # Data streams -> stream infos and data stream states
            for data_stream in trace:
                info = _LttngLiveViewerStreamInfo(
                    self._last_allocated_stream_id,
                    trace_id,
                    False,
                    data_stream.path,
                    data_stream.channel_name,
                )
                trace_stream_infos.append(info)
                self._ds_states[
                    self._last_allocated_stream_id
                ] = _LttngLiveViewerSessionDataStreamState(
                    self, info, data_stream, metadata_stream_id
                )
                self._last_allocated_stream_id += 1

            if trace.creation_timestamp == 0:
                # Only announce streams for traces that are created at
                # the origin.
                #
                # The rest of the streams will be discovered by the
                # client as indexes are received with a "has new data
                # streams" flag set in the reply.
                self._client_visible_stream_infos.extend(trace_stream_infos)

        self._is_attached = False
        fmt = 'Built tracing session state: id={}, name="{}"'
        logging.info(fmt.format(tc_descr.info.tracing_session_id, tc_descr.info.name))

    @property
    def tracing_session_descriptor(self):
        return self._tc_descr

    @property
    def data_stream_states(self):
        return self._ds_states

    @property
    def metadata_stream_states(self):
        return self._ms_states

    @property
    def client_visible_stream_infos(self):
        return self._client_visible_stream_infos

    @property
    def has_new_metadata(self):
        return any(
            [
                ms.is_announced and not ms.all_data_is_sent
                for ms in self._ms_states.values()
            ]
        )

    @property
    def is_attached(self):
        return self._is_attached

    @is_attached.setter
    def is_attached(self, value: bool):
        self._is_attached = value


def needs_new_metadata_section(
    metadata_stream_state: _LttngLiveViewerSessionMetadataStreamState,
    latest_timestamp: int,
):
    if metadata_stream_state.next_section_timestamp is None:
        return False

    if latest_timestamp >= metadata_stream_state.next_section_timestamp:
        return True
    else:
        return False


# An LTTng live viewer session manages a view on tracing sessions
# and replies to commands accordingly.
class _LttngLiveViewerSession:
    def __init__(
        self,
        viewer_session_id: int,
        tracing_session_descriptors: Iterable[LttngTracingSessionDescriptor],
        max_query_data_response_size: Optional[int],
    ):
        self._viewer_session_id = viewer_session_id
        self._ts_states = (
            {}
        )  # type:  dict[int, _LttngLiveViewerSessionTracingSessionState]
        self._stream_states = (
            {}
        )  # type: dict[int, _LttngLiveViewerSessionDataStreamState | _LttngLiveViewerSessionMetadataStreamState]
        self._max_query_data_response_size = max_query_data_response_size
        total_stream_infos = 0

        for ts_descr in tracing_session_descriptors:
            ts_state = _LttngLiveViewerSessionTracingSessionState(
                ts_descr, total_stream_infos
            )
            ts_id = ts_state.tracing_session_descriptor.info.tracing_session_id
            self._ts_states[ts_id] = ts_state
            total_stream_infos += len(ts_state.client_visible_stream_infos)

            # Update session's stream states to have the new states
            self._stream_states.update(ts_state.data_stream_states)
            self._stream_states.update(ts_state.metadata_stream_states)

        self._command_handlers = {
            _LttngLiveViewerAttachToTracingSessionCommand: self._handle_attach_to_tracing_session_command,
            _LttngLiveViewerCreateViewerSessionCommand: self._handle_create_viewer_session_command,
            _LttngLiveViewerDetachFromTracingSessionCommand: self._handle_detach_from_tracing_session_command,
            _LttngLiveViewerGetDataStreamPacketDataCommand: self._handle_get_data_stream_packet_data_command,
            _LttngLiveViewerGetMetadataStreamDataCommand: self._handle_get_metadata_stream_data_command,
            _LttngLiveViewerGetNewStreamInfosCommand: self._handle_get_new_stream_infos_command,
            _LttngLiveViewerGetNextDataStreamIndexEntryCommand: self._handle_get_next_data_stream_index_entry_command,
            _LttngLiveViewerGetTracingSessionInfosCommand: self._handle_get_tracing_session_infos_command,
        }  # type: dict[type[_LttngLiveViewerCommand], Callable[[Any], _LttngLiveViewerReply]]

    @property
    def viewer_session_id(self):
        return self._viewer_session_id

    def _get_tracing_session_state(self, tracing_session_id: int):
        if tracing_session_id not in self._ts_states:
            raise RuntimeError(
                "Unknown tracing session ID {}".format(tracing_session_id)
            )

        return self._ts_states[tracing_session_id]

    def _get_data_stream_state(self, stream_id: int):
        if stream_id not in self._stream_states:
            RuntimeError("Unknown stream ID {}".format(stream_id))

        stream = self._stream_states[stream_id]
        if type(stream) is not _LttngLiveViewerSessionDataStreamState:
            raise RuntimeError("Stream is not a data stream")

        return stream

    def _get_metadata_stream_state(self, stream_id: int):
        if stream_id not in self._stream_states:
            RuntimeError("Unknown stream ID {}".format(stream_id))

        stream = self._stream_states[stream_id]
        if type(stream) is not _LttngLiveViewerSessionMetadataStreamState:
            raise RuntimeError("Stream is not a metadata stream")

        return stream

    def handle_command(self, cmd: _LttngLiveViewerCommand):
        logging.info(
            "Handling command in viewer session: cmd-cls-name={}".format(
                cmd.__class__.__name__
            )
        )
        cmd_type = type(cmd)

        if cmd_type not in self._command_handlers:
            raise RuntimeError(
                "Unexpected command: cmd-cls-name={}".format(cmd.__class__.__name__)
            )

        return self._command_handlers[cmd_type](cmd)

    def _handle_attach_to_tracing_session_command(
        self, cmd: _LttngLiveViewerAttachToTracingSessionCommand
    ):
        fmt = 'Handling "attach to tracing session" command: ts-id={}, offset={}, seek-type={}'
        logging.info(fmt.format(cmd.tracing_session_id, cmd.offset, cmd.seek_type))
        ts_state = self._get_tracing_session_state(cmd.tracing_session_id)
        info = ts_state.tracing_session_descriptor.info

        if ts_state.is_attached:
            raise RuntimeError(
                "Cannot attach to tracing session `{}`: viewer is already attached".format(
                    info.name
                )
            )

        ts_state.is_attached = True
        status = _LttngLiveViewerAttachToTracingSessionReply.Status.OK
        stream_infos_to_announce = ts_state.client_visible_stream_infos

        # Mark stream infos transmitted as part of the reply as
        # announced.
        for si in stream_infos_to_announce:
            if si.is_metadata:
                self._get_metadata_stream_state(si.id).mark_as_announced()
            else:
                self._get_data_stream_state(si.id).mark_as_announced()

        return _LttngLiveViewerAttachToTracingSessionReply(
            status, stream_infos_to_announce
        )

    def _handle_detach_from_tracing_session_command(
        self, cmd: _LttngLiveViewerDetachFromTracingSessionCommand
    ):
        fmt = 'Handling "detach from tracing session" command: ts-id={}'
        logging.info(fmt.format(cmd.tracing_session_id))
        ts_state = self._get_tracing_session_state(cmd.tracing_session_id)
        info = ts_state.tracing_session_descriptor.info

        if not ts_state.is_attached:
            raise RuntimeError(
                "Cannot detach to tracing session `{}`: viewer is not attached".format(
                    info.name
                )
            )

        ts_state.is_attached = False
        status = _LttngLiveViewerDetachFromTracingSessionReply.Status.OK
        return _LttngLiveViewerDetachFromTracingSessionReply(status)

    @staticmethod
    def _stream_is_ready(
        stream_state: _LttngLiveViewerSessionStreamState, creation_timestamp: int
    ):
        return (
            not stream_state.is_announced
            and stream_state.stream.creation_timestamp <= creation_timestamp
        )

    def _needs_new_streams(self, current_timestamp: int):
        return any(
            self._stream_is_ready(ss, current_timestamp)
            for ss in self._stream_states.values()
        )

    def _handle_get_next_data_stream_index_entry_command(
        self, cmd: _LttngLiveViewerGetNextDataStreamIndexEntryCommand
    ):
        fmt = 'Handling "get next data stream index entry" command: stream-id={}'
        logging.info(fmt.format(cmd.stream_id))
        stream_state = self._get_data_stream_state(cmd.stream_id)
        metadata_stream_state = self._get_metadata_stream_state(
            stream_state.metadata_stream_id
        )

        if stream_state.cur_index_entry is None:
            # The viewer is done reading this stream
            status = _LttngLiveViewerGetNextDataStreamIndexEntryReply.Status.HUP

            # Dummy data stream index entry to use with the `HUP` status
            # (the reply needs one, but the viewer ignores it)
            index_entry = _LttngDataStreamIndexEntry(0, 0, 0, 0, 0, 0, 0)

            return _LttngLiveViewerGetNextDataStreamIndexEntryReply(
                status, index_entry, False, False
            )

        timestamp_begin = _get_entry_timestamp_begin(stream_state.cur_index_entry)

        if needs_new_metadata_section(metadata_stream_state, timestamp_begin):
            metadata_stream_state.all_data_is_sent = False
            metadata_stream_state.goto_next_section()

        # The viewer only checks the `has_new_metadata` flag if the
        # reply's status is `OK`, so we need to provide an index here
        has_new_metadata = stream_state.tracing_session_state.has_new_metadata
        if isinstance(stream_state.cur_index_entry, _LttngDataStreamIndexEntry):
            status = _LttngLiveViewerGetNextDataStreamIndexEntryReply.Status.OK
        else:
            status = _LttngLiveViewerGetNextDataStreamIndexEntryReply.Status.INACTIVE

        reply = _LttngLiveViewerGetNextDataStreamIndexEntryReply(
            status,
            stream_state.cur_index_entry,
            has_new_metadata,
            self._needs_new_streams(timestamp_begin),
        )
        self._last_delivered_index_timestamp_begin = timestamp_begin
        stream_state.goto_next_index_entry()
        return reply

    def _handle_get_data_stream_packet_data_command(
        self, cmd: _LttngLiveViewerGetDataStreamPacketDataCommand
    ):
        fmt = 'Handling "get data stream packet data" command: stream-id={}, offset={}, req-length={}'
        logging.info(fmt.format(cmd.stream_id, cmd.offset, cmd.req_length))
        stream_state = self._get_data_stream_state(cmd.stream_id)
        data_response_length = cmd.req_length

        if stream_state.tracing_session_state.has_new_metadata:
            status = _LttngLiveViewerGetDataStreamPacketDataReply.Status.ERROR
            return _LttngLiveViewerGetDataStreamPacketDataReply(
                status, bytes(), True, False
            )

        if self._max_query_data_response_size:
            # Enforce a server side limit on the query requested length.
            # To ensure that the transaction terminate take the minimum of both
            # value.
            data_response_length = min(
                cmd.req_length, self._max_query_data_response_size
            )
            fmt = 'Limiting "get data stream packet data" command: req-length={} actual response size={}'
            logging.info(fmt.format(cmd.req_length, data_response_length))

        data = stream_state.stream.get_data(cmd.offset, data_response_length)
        status = _LttngLiveViewerGetDataStreamPacketDataReply.Status.OK
        return _LttngLiveViewerGetDataStreamPacketDataReply(status, data, False, False)

    def _handle_get_metadata_stream_data_command(
        self, cmd: _LttngLiveViewerGetMetadataStreamDataCommand
    ):
        fmt = 'Handling "get metadata stream data" command: stream-id={}'
        logging.info(fmt.format(cmd.stream_id))
        metadata_stream_state = self._get_metadata_stream_state(cmd.stream_id)

        if metadata_stream_state.all_data_is_sent:
            status = _LttngLiveViewerGetMetadataStreamDataContentReply.Status.NO_NEW
            return _LttngLiveViewerGetMetadataStreamDataContentReply(status, bytes())

        metadata_stream_state.all_data_is_sent = True
        status = _LttngLiveViewerGetMetadataStreamDataContentReply.Status.OK
        metadata_section = metadata_stream_state.cur_section
        assert metadata_section is not None

        # If we are sending an empty section, ready the next one right away.
        if len(metadata_section.data) == 0:
            metadata_stream_state.all_data_is_sent = False
            metadata_stream_state.goto_next_section()

        fmt = 'Replying to "get metadata stream data" command: metadata-size={}'
        logging.info(fmt.format(len(metadata_section.data)))
        return _LttngLiveViewerGetMetadataStreamDataContentReply(
            status, metadata_section.data
        )

    def _get_stream_infos_ready_for_announcement(self):
        ready_stream_infos = []  # type: list[_LttngLiveViewerStreamInfo]

        for ss in self._stream_states.values():
            if self._stream_is_ready(ss, ss.stream.creation_timestamp):
                ready_stream_infos.append(ss.info)

        return ready_stream_infos

    # A stream is considered finished if it has been announced and all
    # of its index entries have been provided to the client.
    def _all_streams_finished(self):
        return all(
            isinstance(stream_state, _LttngLiveViewerSessionMetadataStreamState)
            or (stream_state.cur_index_entry is None and stream_state.is_announced)
            for stream_state in self._stream_states.values()
        )

    def _handle_get_new_stream_infos_command(
        self, cmd: _LttngLiveViewerGetNewStreamInfosCommand
    ):
        fmt = 'Handling "get new stream infos" command: ts-id={}'
        logging.info(fmt.format(cmd.tracing_session_id))
        newly_announced_stream_infos = self._get_stream_infos_ready_for_announcement()

        # Mark stream infos transmitted as part of the reply as
        # announced.
        for si in newly_announced_stream_infos:
            if si.is_metadata:
                self._get_metadata_stream_state(si.id).mark_as_announced()
            else:
                self._get_data_stream_state(si.id).mark_as_announced()

        status = _LttngLiveViewerGetNewStreamInfosReply.Status.OK

        if len(newly_announced_stream_infos) == 0:
            # If all streams have been transmitted and no new traces are
            # scheduled for creation, hang up to signal that the tracing
            # session is "done".
            status = (
                _LttngLiveViewerGetNewStreamInfosReply.Status.HUP
                if self._all_streams_finished()
                else _LttngLiveViewerGetNewStreamInfosReply.Status.NO_NEW
            )

        return _LttngLiveViewerGetNewStreamInfosReply(
            status, newly_announced_stream_infos
        )

    def _handle_get_tracing_session_infos_command(
        self, cmd: _LttngLiveViewerGetTracingSessionInfosCommand
    ):
        logging.info('Handling "get tracing session infos" command.')
        infos = [
            tss.tracing_session_descriptor.info for tss in self._ts_states.values()
        ]
        infos.sort(key=lambda info: info.name)
        return _LttngLiveViewerGetTracingSessionInfosReply(infos)

    def _handle_create_viewer_session_command(
        self, cmd: _LttngLiveViewerCreateViewerSessionCommand
    ):
        logging.info('Handling "create viewer session" command.')
        status = _LttngLiveViewerCreateViewerSessionReply.Status.OK

        # This does nothing here. In the LTTng relay daemon, it
        # allocates the viewer session's state.
        return _LttngLiveViewerCreateViewerSessionReply(status)


# An LTTng live TCP server.
#
# On creation, it binds to `localhost` on the TCP port `port` if not `None`, or
# on an OS-assigned TCP port otherwise. It writes the decimal TCP port number
# to a temporary port file.  It renames the temporary port file to
# `port_filename`.
#
# `tracing_session_descriptors` is a list of tracing session descriptors
# (`LttngTracingSessionDescriptor`) to serve.
#
# This server accepts a single viewer (client).
#
# When the viewer closes the connection, the server's constructor
# returns.
class LttngLiveServer:
    def __init__(
        self,
        port: Optional[int],
        port_filename: Optional[str],
        tracing_session_descriptors: Iterable[LttngTracingSessionDescriptor],
        max_query_data_response_size: Optional[int],
        max_minor_version: int,
    ):
        logging.info("Server configuration:")

        logging.info("  Maximum minor version: {}".format(max_minor_version))

        if port_filename is not None:
            logging.info("  Port file name: `{}`".format(port_filename))

        if max_query_data_response_size is not None:
            logging.info(
                "  Maximum response data query size: `{}`".format(
                    max_query_data_response_size
                )
            )

        for ts_descr in tracing_session_descriptors:
            info = ts_descr.info
            fmt = '  TS descriptor: name="{}", id={}, hostname="{}", live-timer-freq={}, client-count={}, stream-count={}, trace-format={}:'
            logging.info(
                fmt.format(
                    info.name,
                    info.tracing_session_id,
                    info.hostname,
                    info.live_timer_freq,
                    info.client_count,
                    info.stream_count,
                    info.trace_format.name,
                )
            )

            for trace in ts_descr.traces:
                logging.info('    Trace: path="{}"'.format(trace.path))

        self._max_minor_version = max_minor_version
        self._ts_descriptors = tracing_session_descriptors
        self._max_query_data_response_size = max_query_data_response_size
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._codec = _LttngLiveViewerProtocolCodec()

        # Port 0: OS assigns an unused port
        serv_addr = ("localhost", port if port is not None else 0)
        self._sock.bind(serv_addr)

        if port_filename is not None:
            self._write_port_to_file(port_filename)

        print("Listening on port {}".format(self._server_port))

        for ts_descr in tracing_session_descriptors:
            info = ts_descr.info
            print(
                "net://localhost:{}/host/{}/{}".format(
                    self._server_port, info.hostname, info.name
                )
            )

        try:
            self._listen()
        finally:
            self._sock.close()
            logging.info("Closed connection and socket.")

    @property
    def _server_port(self):
        return self._sock.getsockname()[1]

    def _recv_command(self):
        data = bytes()

        while True:
            logging.info("Waiting for viewer command.")
            buf = self._conn.recv(128)

            if not buf:
                logging.info("Client closed connection.")

                if data:
                    raise RuntimeError(
                        "Client closed connection after having sent {} command bytes.".format(
                            len(data)
                        )
                    )

                return

            logging.info("Received data from viewer: length={}".format(len(buf)))

            data += buf

            try:
                cmd = self._codec.decode(data)
            except struct.error as exc:
                raise RuntimeError("Malformed command: {}".format(exc)) from exc

            if cmd is not None:
                logging.info(
                    "Received command from viewer: cmd-cls-name={}".format(
                        cmd.__class__.__name__
                    )
                )
                return cmd

    def _send_reply(self, reply: _LttngLiveViewerReply):
        data = self._codec.encode(reply)
        logging.info(
            "Sending reply to viewer: reply-cls-name={}, length={}".format(
                reply.__class__.__name__, len(data)
            )
        )
        self._conn.sendall(data)

    def _handle_connection(self):
        # First command must be "connect"
        cmd = self._recv_command()

        if type(cmd) is not _LttngLiveViewerConnectCommand:
            raise RuntimeError(
                'First command is not "connect": cmd-cls-name={}'.format(
                    cmd.__class__.__name__
                )
            )

        # Create viewer session (arbitrary ID 23)
        logging.info(
            "LTTng live viewer connected: version={}.{}".format(cmd.major, cmd.minor)
        )
        viewer_session = _LttngLiveViewerSession(
            23, self._ts_descriptors, self._max_query_data_response_size
        )

        # Set our effective minor version
        self._minor_version = min([self._max_minor_version, cmd.minor])
        logging.info(
            "Effective server version is available: version={}.{}".format(
                2, self._minor_version
            )
        )
        self._codec.server_minor_version = self._minor_version

        # Send "connect" reply
        self._send_reply(
            _LttngLiveViewerConnectReply(
                viewer_session.viewer_session_id, 2, self._minor_version
            )
        )

        # Make the viewer session handle the remaining commands
        while True:
            cmd = self._recv_command()

            if cmd is None:
                # Connection closed (at an expected location within the
                # conversation)
                return

            self._send_reply(viewer_session.handle_command(cmd))

    def _listen(self):
        logging.info("Listening: port={}".format(self._server_port))
        # Backlog must be present for Python version < 3.5.
        # 128 is an arbitrary number since we expect only 1 connection anyway.
        self._sock.listen(128)
        self._conn, viewer_addr = self._sock.accept()
        logging.info(
            "Accepted viewer: addr={}:{}".format(viewer_addr[0], viewer_addr[1])
        )

        try:
            self._handle_connection()
        finally:
            self._conn.close()

    def _write_port_to_file(self, port_filename: str):
        # Write the port number to a temporary file.
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, dir=os.path.dirname(port_filename)
        ) as tmp_port_file:
            print(self._server_port, end="", file=tmp_port_file)

        # Rename temporary file to real file.
        #
        # For unknown reasons, on Windows, moving the port file from its
        # temporary location to its final location (where the user of
        # the server expects it to appear) may raise a `PermissionError`
        # exception.
        #
        # We suppose it's possible that something in the Windows kernel
        # hasn't completely finished using the file when we try to move
        # it.
        #
        # Use a wait-and-retry scheme as a (bad) workaround.
        num_attempts = 5
        retry_delay_s = 1

        for attempt in reversed(range(num_attempts)):
            try:
                os.replace(tmp_port_file.name, port_filename)
                logging.info(
                    'Renamed port file: src-path="{}", dst-path="{}"'.format(
                        tmp_port_file.name, port_filename
                    )
                )
                return
            except PermissionError:
                logging.info(
                    'Permission error while attempting to rename port file; retrying in {} second: src-path="{}", dst-path="{}"'.format(
                        retry_delay_s, tmp_port_file.name, port_filename
                    )
                )

                if attempt == 0:
                    raise

                time.sleep(retry_delay_s)


def _session_descriptors_from_path(
    sessions_filename: str, trace_path_prefix: Optional[str]
):
    # File format is:
    #
    #     [
    #         {
    #             "name": "my-session",
    #             "id": 17,
    #             "hostname": "myhost",
    #             "live-timer-freq": 1000000,
    #             "client-count": 23,
    #             "traces": [
    #                 {
    #                     "path": "lol",
    #                     creation-timestamp: 12948
    #                 },
    #                 {
    #                     "path": "meow/mix",
    #                     "beacons": {
    #                         "my_stream": [ 5235787, 728375283 ]
    #                     },
    #                     "metadata-sections": [
    #                           {
    #                                "line": 1,
    #                                "timestamp": 0
    #                           }
    #                      ]
    #                 }
    #             ]
    #         }
    #     ]
    with open(sessions_filename, "r") as sessions_file:
        sessions_json = tjson.load(sessions_file, tjson.ArrayVal)

    sessions = []  # type: list[LttngTracingSessionDescriptor]

    for session_json in sessions_json.iter(tjson.ObjVal):
        name = session_json.at("name", tjson.StrVal).val
        tracing_session_id = session_json.at("id", tjson.IntVal).val
        hostname = session_json.at("hostname", tjson.StrVal).val
        live_timer_freq = session_json.at("live-timer-freq", tjson.IntVal).val
        client_count = session_json.at("client-count", tjson.IntVal).val
        traces_json = session_json.at("traces", tjson.ArrayVal)

        traces = []  # type: list[LttngTrace]

        for trace_json in traces_json.iter(tjson.ObjVal):
            metadata_sections = (
                trace_json.at("metadata-sections", tjson.ArrayVal)
                if "metadata-sections" in trace_json
                else None
            )
            beacons = (
                trace_json.at("beacons", tjson.ObjVal)
                if "beacons" in trace_json
                else None
            )
            path = trace_json.at("path", tjson.StrVal).val
            creation_timestamp = (
                trace_json.at("creation-timestamp", tjson.IntVal).val
                if "creation-timestamp" in trace_json
                else 0
            )  # type: int

            if not os.path.isabs(path) and trace_path_prefix:
                path = os.path.join(trace_path_prefix, path)

            traces.append(
                LttngTrace(path, metadata_sections, beacons, creation_timestamp)
            )

        assert len(traces) > 0

        sessions.append(
            LttngTracingSessionDescriptor(
                name,
                tracing_session_id,
                hostname,
                live_timer_freq,
                client_count,
                traces[0].metadata_stream.trace_format,
                traces,
            )
        )

    return sessions


def _loglevel_parser(string: str):
    loglevels = {"info": logging.INFO, "warning": logging.WARNING}
    if string not in loglevels:
        msg = "{} is not a valid loglevel".format(string)
        raise argparse.ArgumentTypeError(msg)
    return loglevels[string]


if __name__ == "__main__":
    logging.basicConfig(format="# %(asctime)-25s%(message)s")
    parser = argparse.ArgumentParser(
        description="LTTng-live protocol mocker", add_help=False
    )
    parser.add_argument(
        "--log-level",
        default="warning",
        choices=["info", "warning"],
        help="The loglevel to be used.",
    )

    loglevel_namespace, remaining_args = parser.parse_known_args()
    logging.getLogger().setLevel(_loglevel_parser(loglevel_namespace.log_level))

    parser.add_argument(
        "--port",
        help="The port to bind to. If missing, use an OS-assigned port.",
        type=int,
    )
    parser.add_argument(
        "--port-filename",
        help="The final port file. This file is present when the server is ready to receive connection.",
    )
    parser.add_argument(
        "--max-query-data-response-size",
        type=int,
        help="The maximum size of control data response in bytes.",
    )
    parser.add_argument(
        "--trace-path-prefix",
        type=str,
        help="Prefix to prepend to the trace paths of session configurations.",
    )
    parser.add_argument(
        "--server-max-minor-version",
        type=int,
        default=10,
        help="Maximum minor version of the server instead of 10.",
    )
    parser.add_argument(
        "sessions_filename",
        type=str,
        help="Path to a session configuration file.",
        metavar="sessions-filename",
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    args = parser.parse_args(args=remaining_args)
    sessions_filename = args.sessions_filename  # type: str
    trace_path_prefix = args.trace_path_prefix  # type: str | None
    sessions = _session_descriptors_from_path(
        sessions_filename,
        trace_path_prefix,
    )

    port = args.port  # type: int | None
    port_filename = args.port_filename  # type: str | None
    max_query_data_response_size = args.max_query_data_response_size  # type: int | None
    server_max_minor_version = args.server_max_minor_version  # type: int
    LttngLiveServer(
        port,
        port_filename,
        sessions,
        max_query_data_response_size,
        server_max_minor_version,
    )



# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (C) 2019 EfficiOS Inc.
#

import unittest

import bt2
from bt2 import component as bt2_component


class UserComponentTestCase(unittest.TestCase):
    @staticmethod
    def _create_comp(comp_cls, name=None, log_level=bt2.LoggingLevel.NONE):
        graph = bt2.Graph()

        if name is None:
            name = "comp"

        return graph.add_component(comp_cls, name, logging_level=log_level)

    def test_name(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self.name, "yaes")

            def _user_consume(self):
                pass

        self._create_comp(MySink, "yaes")

    def test_logging_level(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self.logging_level, bt2.LoggingLevel.INFO)

            def _user_consume(self):
                pass

        self._create_comp(MySink, "yaes", bt2.LoggingLevel.INFO)

    def test_graph_mip_version(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self._graph_mip_version, 0)

            def _user_consume(self):
                pass

        self._create_comp(MySink, "yaes", bt2.LoggingLevel.INFO)

    def test_class(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self.cls, MySink)

            def _user_consume(self):
                pass

        self._create_comp(MySink)

    def test_addr(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertIsInstance(comp_self.addr, int)
                self.assertNotEqual(comp_self.addr, 0)

            def _user_consume(self):
                pass

        self._create_comp(MySink)

    def test_finalize(self):
        finalized = False

        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            def _user_finalize(comp_self):
                nonlocal finalized
                finalized = True

        graph = bt2.Graph()
        comp = graph.add_component(MySink, "lel")

        del graph
        del comp
        self.assertTrue(finalized)

    def test_source_component_config(self):
        class MySource(
            bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            def __init__(comp_self, config, params, obj):
                nonlocal cfg_type
                cfg_type = type(config)

        cfg_type = None
        self._create_comp(MySource)
        self.assertIs(cfg_type, bt2_component._UserSourceComponentConfiguration)

    def test_filter_component_config(self):
        class MyFilter(
            bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            def __init__(comp_self, config, params, obj):
                nonlocal cfg_type
                cfg_type = type(config)

        cfg_type = None
        self._create_comp(MyFilter)
        self.assertIs(cfg_type, bt2_component._UserFilterComponentConfiguration)

    def test_sink_component_config(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                nonlocal cfg_type
                cfg_type = type(config)

            def _user_consume(self):
                pass

        cfg_type = None
        self._create_comp(MySink)
        self.assertIs(cfg_type, bt2_component._UserSinkComponentConfiguration)


class GenericComponentTestCase(unittest.TestCase):
    @staticmethod
    def _create_comp(comp_cls, name=None, log_level=bt2.LoggingLevel.NONE):
        graph = bt2.Graph()

        if name is None:
            name = "comp"

        return graph.add_component(comp_cls, name, logging_level=log_level)

    def test_name(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink, "yaes")
        self.assertEqual(comp.name, "yaes")

    def test_logging_level(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink, "yaes", bt2.LoggingLevel.WARNING)
        self.assertEqual(comp.logging_level, bt2.LoggingLevel.WARNING)

    def test_class(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink)
        self.assertEqual(comp.cls, MySink)

    def test_addr(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink)
        self.assertIsInstance(comp.addr, int)
        self.assertNotEqual(comp.addr, 0)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (C) 2019 EfficiOS Inc.
#

# pyright: strict, reportPrivateUsage=false

import math
import typing

import bt2

bt2.register_plugin(__name__, "trace-ir-test")

_FieldType = typing.TypeVar("_FieldType", bound=bt2._Field)


def c(field: bt2._Field, expected_type: typing.Type[_FieldType]) -> _FieldType:
    assert type(field) is expected_type
    return field


class AllFieldsIter(bt2._UserMessageIterator):
    def __init__(
        self,
        config: bt2._MessageIteratorConfiguration,
        output_port: bt2._UserComponentOutputPort,
    ):
        mip = self._component._graph_mip_version
        ec = typing.cast(bt2._EventClass, output_port.user_data)
        sc = ec.stream_class
        tc = sc.trace_class
        stream = tc().create_stream(sc)
        ev = self._create_event_message(ec, stream)

        payload = ev.event.payload_field
        assert payload

        payload["bool"] = False
        payload["real_single"] = 2.0
        payload["real_double"] = math.pi

        # Bit array (with flags if MIP > 0)
        c(payload["bit-array"], bt2._BitArrayField).value_as_integer = 0b1000000

        payload["int32"] = 121
        payload["int3"] = -1
        payload["int9_hex"] = -92
        payload["uint32"] = 121
        payload["uint61"] = 299792458
        payload["uint5_oct"] = 29

        # Struct
        struct = c(payload["struct"], bt2._StructureField)
        struct["str"] = "Rotisserie St-Hubert"
        struct["option_real"] = math.pi

        payload["string"] = "🎉"

        # Static array
        payload["static-array"] = ["🕰", "🦴", " 🎍"]

        # Dynamic array without length field
        payload["dynamic-array-without-length-field"] = [1.2, 2 / 3, 42.3, math.pi]

        # Dynamic array with length field
        payload["dynamic-array-with-length-field-length"] = 4
        payload["dynamic-array-with-length-field"] = [5.2, 5 / 3, 42.5, math.pi * 12]

        # Option without selector field
        payload["option-without-selector-field-with-value"] = "NORMANDIN"

        # Options with bool selector field
        payload["option-with-bool-selector-field-selector-true"] = True
        payload["option-with-bool-selector-field-selector-false"] = False
        payload["option-with-bool-selector-field-with-value"] = "Mike's"
        payload["option-with-reversed-bool-selector-field-with-value"] = "Boustan"

        # Option with unsigned integer selector field
        payload["option-with-unsigned-integer-selector-field-selector"] = 1
        payload[
            "option-with-unsigned-integer-selector-field"
        ] = "Barbies resto bar grill"

        # Option with signed integer selector field
        payload["option-with-signed-integer-selector-field-selector"] = 1
        payload["option-with-signed-integer-selector-field"] = "Pacini"

        # Variant without selector field
        variant = c(payload["variant-without-selector-field"], bt2._VariantField)
        variant.selected_option_index = 1
        variant.value = "Couche-Tard"

        # Variant with unsigned integer selector field
        c(
            payload["variant-with-unsigned-integer-selector-field-selector"],
            bt2._UnsignedIntegerField,
        ).value = 5
        variant_with_unsigned_integer_selector_field = c(
            payload["variant-with-unsigned-integer-selector-field"],
            bt2._VariantFieldWithUnsignedIntegerSelectorField,
        )
        variant_with_unsigned_integer_selector_field.selected_option_index = 0
        variant_with_unsigned_integer_selector_field.value = "Marco Smith"

        # Variant with signed integer selector field
        c(
            payload["variant-with-signed-integer-selector-field-selector"],
            bt2._SignedIntegerField,
        ).value = 5
        variant_with_signed_integer_selector_field = c(
            payload["variant-with-signed-integer-selector-field"],
            bt2._VariantFieldWithSignedIntegerSelectorField,
        )
        variant_with_signed_integer_selector_field.selected_option_index = 0
        variant_with_signed_integer_selector_field.value = "Pascal Lefebvre"

        if mip > 0:
            # Static blob
            c(payload["static-blob"], bt2._StaticBlobField).data = b"Masbourian"

            # Dynamic blob without length field
            dynamic_blob_without_length_field = c(
                payload["dynamic-blob-without-length-field"], bt2._DynamicBlobField
            )
            dynamic_blob_without_length_field.length = 3
            dynamic_blob_without_length_field.data = b"mdr"

            # Dynamic blob with length field
            c(
                payload["dynamic-blob-with-length-field-length"],
                bt2._UnsignedIntegerField,
            ).value = 5
            dynamic_blob_with_length_field = c(
                payload["dynamic-blob-with-length-field"],
                bt2._DynamicBlobFieldWithLengthField,
            )
            dynamic_blob_with_length_field.length = 5
            dynamic_blob_with_length_field.data = b"buick"

        self._msgs = [
            self._create_stream_beginning_message(stream),
            ev,
            self._create_stream_end_message(stream),
        ]

    def __next__(self):
        if len(self._msgs) > 0:
            return self._msgs.pop(0)
        else:
            raise StopIteration


@bt2.plugin_component_class
class AllFields(bt2._UserSourceComponent, message_iterator_class=AllFieldsIter):
    @staticmethod
    def _user_get_supported_mip_versions(
        params: bt2._MapValueConst, obj: object, log_level: bt2.LoggingLevel
    ):
        return [0, 1]

    def __init__(
        self,
        config: bt2._UserSourceComponentConfiguration,
        params: bt2._MapValueConst,
        obj: object,
    ):
        mip = self._graph_mip_version
        tc = self._create_trace_class()

        # Dynamic array with length field
        dynamic_array_with_length_field_length = tc.create_unsigned_integer_field_class(
            19
        )

        # Options with bool selector field
        option_with_bool_selector_field_selector_true = tc.create_bool_field_class()
        option_with_bool_selector_field_selector_false = tc.create_bool_field_class()
        option_with_bool_selector_field_selector_true_fl = (
            tc.create_field_location(
                bt2.FieldLocationScope.EVENT_PAYLOAD,
                ["option-with-bool-selector-field-selector-true"],
            )
            if mip > 0
            else None
        )
        option_with_bool_selector_field_selector_false_fl = (
            tc.create_field_location(
                bt2.FieldLocationScope.EVENT_PAYLOAD,
                ["option-with-bool-selector-field-selector-false"],
            )
            if mip > 0
            else None
        )

        # Options with integer selector field
        option_with_unsigned_integer_selector_field_selector = (
            tc.create_unsigned_integer_field_class(8)
        )
        option_with_unsigned_integer_selector_field_ranges = (
            bt2.UnsignedIntegerRangeSet([(1, 3), (18, 44)])
        )
        option_with_signed_integer_selector_field_selector = (
            tc.create_signed_integer_field_class(8)
        )
        option_with_signed_integer_selector_field_ranges = bt2.SignedIntegerRangeSet(
            [(1, 3), (18, 44)]
        )

        # Variants with integer selector field
        variant_with_unsigned_selector_field_selector = (
            tc.create_unsigned_integer_field_class(8)
        )
        variant_with_unsigned_selector_field_options = (
            (
                "var_str",
                tc.create_string_field_class(),
                bt2.UnsignedIntegerRangeSet([(5, 5)]),
            ),
        )
        variant_with_signed_selector_field_selector = (
            tc.create_signed_integer_field_class(8)
        )
        variant_with_signed_selector_field_options = (
            (
                "var_str",
                tc.create_string_field_class(),
                bt2.SignedIntegerRangeSet([(5, 5)]),
            ),
        )

        payload = tc.create_structure_field_class(
            members=(
                ("bool", tc.create_bool_field_class()),
                ("real_single", tc.create_single_precision_real_field_class()),
                ("real_double", tc.create_double_precision_real_field_class()),
                (
                    "bit-array",
                    tc.create_bit_array_field_class(
                        16,
                        flags=(
                            (
                                (
                                    "flag1",
                                    bt2.UnsignedIntegerRangeSet(((1, 3), (6, 9))),
                                ),
                                ("flag2", bt2.UnsignedIntegerRangeSet([(4, 6)])),
                            )
                            if mip > 0
                            else None
                        ),
                    ),
                ),
                ("int32", tc.create_signed_integer_field_class(32)),
                ("int3", tc.create_signed_integer_field_class(3)),
                (
                    "int9_hex",
                    tc.create_signed_integer_field_class(
                        9,
                        preferred_display_base=bt2.IntegerDisplayBase.HEXADECIMAL,
                    ),
                ),
                ("uint32", tc.create_unsigned_integer_field_class(32)),
                ("uint61", tc.create_unsigned_integer_field_class(61)),
                (
                    "uint5_oct",
                    tc.create_unsigned_integer_field_class(
                        5, preferred_display_base=bt2.IntegerDisplayBase.OCTAL
                    ),
                ),
                (
                    "struct",
                    tc.create_structure_field_class(
                        members=(
                            ("str", tc.create_string_field_class()),
                            (
                                "option_real",
                                tc.create_option_field_class_without_selector_field(
                                    tc.create_double_precision_real_field_class()
                                ),
                            ),
                        )
                    ),
                ),
                ("string", tc.create_string_field_class()),
                (
                    "dynamic-array-without-length-field",
                    tc.create_dynamic_array_field_class(
                        tc.create_double_precision_real_field_class()
                    ),
                ),
                (
                    "dynamic-array-with-length-field-length",
                    dynamic_array_with_length_field_length,
                ),
                (
                    "dynamic-array-with-length-field",
                    (
                        tc.create_dynamic_array_field_class(
                            tc.create_double_precision_real_field_class(),
                            length_fc=dynamic_array_with_length_field_length,
                        )
                        if mip == 0
                        else tc.create_dynamic_array_field_class(
                            tc.create_double_precision_real_field_class(),
                            length_field_location=tc.create_field_location(
                                bt2.FieldLocationScope.EVENT_PAYLOAD, ["dyn_array_len"]
                            ),
                        )
                    ),
                ),
                (
                    "static-array",
                    tc.create_static_array_field_class(
                        tc.create_string_field_class(), 3
                    ),
                ),
                (
                    "option-without-selector-field-without-value",
                    tc.create_option_field_class_without_selector_field(
                        tc.create_double_precision_real_field_class()
                    ),
                ),
                (
                    "option-without-selector-field-with-value",
                    tc.create_option_field_class_without_selector_field(
                        tc.create_string_field_class()
                    ),
                ),
                (
                    "option-with-bool-selector-field-selector-true",
                    option_with_bool_selector_field_selector_true,
                ),
                (
                    "option-with-bool-selector-field-selector-false",
                    option_with_bool_selector_field_selector_false,
                ),
                (
                    "option-with-bool-selector-field-with-value",
                    (
                        (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_fc=option_with_bool_selector_field_selector_true,
                            )
                        )
                        if mip == 0
                        else (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_field_location=option_with_bool_selector_field_selector_true_fl,
                            )
                        )
                    ),
                ),
                (
                    "option-with-bool-selector-field-without-value",
                    (
                        (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_fc=option_with_bool_selector_field_selector_false,
                            )
                        )
                        if mip == 0
                        else (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_field_location=option_with_bool_selector_field_selector_false_fl,
                            )
                        )
                    ),
                ),
                (
                    "option-with-reversed-bool-selector-field-with-value",
                    (
                        (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_fc=option_with_bool_selector_field_selector_false,
                                selector_is_reversed=True,
                            )
                        )
                        if mip == 0
                        else (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_field_location=option_with_bool_selector_field_selector_false_fl,
                                selector_is_reversed=True,
                            )
                        )
                    ),
                ),
                (
                    "option-with-reversed-bool-selector-field-without-value",
                    (
                        (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_fc=option_with_bool_selector_field_selector_true,
                                selector_is_reversed=True,
                            )
                        )
                        if mip == 0
                        else (
                            tc.create_option_field_class_with_bool_selector_field(
                                tc.create_string_field_class(),
                                selector_field_location=option_with_bool_selector_field_selector_true_fl,
                                selector_is_reversed=True,
                            )
                        )
                    ),
                ),
                (
                    "option-with-unsigned-integer-selector-field-selector",
                    option_with_unsigned_integer_selector_field_selector,
                ),
                (
                    "option-with-unsigned-integer-selector-field",
                    (
                        (
                            tc.create_option_field_class_with_integer_selector_field(
                                tc.create_string_field_class(),
                                selector_fc=option_with_unsigned_integer_selector_field_selector,
                                ranges=option_with_unsigned_integer_selector_field_ranges,
                            )
                        )
                        if mip == 0
                        else (
                            tc.create_option_field_class_with_unsigned_integer_selector_field(
                                tc.create_string_field_class(),
                                selector_field_location=tc.create_field_location(
                                    bt2.FieldLocationScope.EVENT_PAYLOAD,
                                    [
                                        "option-with-unsigned-integer-selector-field-selector"
                                    ],
                                ),
                                ranges=option_with_unsigned_integer_selector_field_ranges,
                            )
                        )
                    ),
                ),
                (
                    "option-with-signed-integer-selector-field-selector",
                    option_with_signed_integer_selector_field_selector,
                ),
                (
                    "option-with-signed-integer-selector-field",
                    (
                        (
                            tc.create_option_field_class_with_integer_selector_field(
                                tc.create_string_field_class(),
                                selector_fc=option_with_signed_integer_selector_field_selector,
                                ranges=option_with_signed_integer_selector_field_ranges,
                            )
                        )
                        if mip == 0
                        else (
                            tc.create_option_field_class_with_signed_integer_selector_field(
                                tc.create_string_field_class(),
                                selector_field_location=tc.create_field_location(
                                    bt2.FieldLocationScope.EVENT_PAYLOAD,
                                    [
                                        "option-with-signed-integer-selector-field-selector"
                                    ],
                                ),
                                ranges=option_with_signed_integer_selector_field_ranges,
                            )
                        )
                    ),
                ),
                (
                    "variant-without-selector-field",
                    (
                        tc.create_variant_field_class
                        if mip == 0
                        else tc.create_variant_field_class_without_selector_field
                    )(
                        options=(
                            ("opt1", tc.create_string_field_class()),
                            (
                                "opt2" if mip == 0 else None,
                                tc.create_string_field_class(),
                            ),
                        )
                    ),
                ),
                (
                    "variant-with-unsigned-integer-selector-field-selector",
                    variant_with_unsigned_selector_field_selector,
                ),
                (
                    "variant-with-unsigned-integer-selector-field",
                    (
                        tc.create_variant_field_class(
                            selector_fc=variant_with_unsigned_selector_field_selector,
                            options=variant_with_unsigned_selector_field_options,
                        )
                        if mip == 0
                        else (
                            tc.create_variant_field_class_with_unsigned_integer_selector_field(
                                selector_field_location=tc.create_field_location(
                                    bt2.FieldLocationScope.EVENT_PAYLOAD,
                                    [
                                        "variant-with-unsigned-integer-selector-field-selector"
                                    ],
                                ),
                                options=variant_with_unsigned_selector_field_options,
                            )
                        )
                    ),
                ),
                (
                    "variant-with-signed-integer-selector-field-selector",
                    variant_with_signed_selector_field_selector,
                ),
                (
                    "variant-with-signed-integer-selector-field",
                    (
                        tc.create_variant_field_class(
                            selector_fc=variant_with_signed_selector_field_selector,
                            options=variant_with_signed_selector_field_options,
                        )
                        if mip == 0
                        else (
                            tc.create_variant_field_class_with_signed_integer_selector_field(
                                selector_field_location=tc.create_field_location(
                                    bt2.FieldLocationScope.EVENT_PAYLOAD,
                                    [
                                        "variant-with-signed-integer-selector-field-selector"
                                    ],
                                ),
                                options=variant_with_signed_selector_field_options,
                            )
                        )
                    ),
                ),
            )
        )

        if mip > 0:
            payload += (
                (
                    "static-blob",
                    tc.create_static_blob_field_class(
                        10, media_type="application/x-gameboy-rom"
                    ),
                ),
                (
                    "dynamic-blob-without-length-field",
                    tc.create_dynamic_blob_field_class(
                        media_type="application/x-shockwave-flash"
                    ),
                ),
                (
                    "dynamic-blob-with-length-field-length",
                    tc.create_unsigned_integer_field_class(8),
                ),
                (
                    "dynamic-blob-with-length-field",
                    tc.create_dynamic_blob_field_class(
                        length_field_location=tc.create_field_location(
                            bt2.FieldLocationScope.EVENT_PAYLOAD,
                            ["dynamic-blob-with-length-field-length"],
                        ),
                        media_type="application/vnd.wordperfect",
                    ),
                ),
            )

        self._add_output_port(
            "out",
            tc.create_stream_class().create_event_class(payload_field_class=payload),
        )



# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (C) 2019 EfficiOS Inc.
#

import bt2


@bt2.plugin_component_class
class ThisIsASource(
    bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
):
    pass


@bt2.plugin_component_class
class ThisIsAFilter(
    bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
):
    pass


@bt2.plugin_component_class
class ThisIsASink(bt2._UserSinkComponent):
    def _user_consume(self):
        pass


bt2.register_plugin(
    __name__,
    "this-is-a-plugin",
    version=(1, 2, 3, "bob"),
    description="A plugin",
    author="Jorge Mario Bergoglio",
    license="The license",
)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (C) 2020 Geneviève Bastien <gbastien@versatic.net>

import bt2


class TheIteratorOfProblems(bt2._UserMessageIterator):
    def __init__(self, config, port):
        tc, sc, ec1, params = port.user_data
        trace = tc()
        stream = trace.create_stream(sc)
        event_value = params["value"]
        self._msgs = []

        self._msgs.append(self._create_stream_beginning_message(stream))

        ev_msg1 = self._create_event_message(ec1, stream)
        ev_msg1.event.payload_field["enum_field"] = event_value

        self._msgs.append(ev_msg1)

        self._msgs.append(self._create_stream_end_message(stream))

        self._at = 0
        config.can_seek_forward = True

    def _user_seek_beginning(self):
        self._at = 0

    def __next__(self):
        if self._at < len(self._msgs):
            msg = self._msgs[self._at]
            self._at += 1
            return msg
        else:
            raise StopIteration


@bt2.plugin_component_class
class TheSourceOfProblems(
    bt2._UserSourceComponent, message_iterator_class=TheIteratorOfProblems
):
    def __init__(self, config, params, obj):
        tc = self._create_trace_class()

        enum_values_str = params["enum-values"]

        sc = tc.create_stream_class()

        # Create the enumeration field with the values in parameter
        if params["enum-signed"]:
            enumfc = tc.create_signed_enumeration_field_class()
        else:
            enumfc = tc.create_unsigned_enumeration_field_class()

        groups = str(enum_values_str).split(" ")
        mappings = {}
        range_set_type = (
            bt2.SignedIntegerRangeSet
            if params["enum-signed"]
            else bt2.UnsignedIntegerRangeSet
        )
        for group in groups:
            label, low, high = group.split(",")

            if label not in mappings.keys():
                mappings[label] = range_set_type()

            mappings[label].add((int(low), int(high)))

        for x, y in mappings.items():
            enumfc.add_mapping(x, y)

        self._add_output_port(
            "out",
            (
                tc,
                sc,
                sc.create_event_class(
                    name="with_enum",
                    payload_field_class=tc.create_structure_field_class(
                        members=(("enum_field", enumfc),)
                    ),
                ),
                params,
            ),
        )


bt2.register_plugin(__name__, "test-pretty")



# SPDX-License-Identifier: MIT
#
# Copyright (c) 2020 Philippe Proulx <pproulx@efficios.com>

# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import os
import re
import signal
import os.path
import unittest
import functools
import subprocess

import tjson

# the `conds-triggers` program's full path
_CONDS_TRIGGERS_PATH = os.environ["BT_TESTS_LIB_CONDS_TRIGGER_BIN"]


# test methods are added by _create_tests()
class LibPrePostCondsTestCase(unittest.TestCase):
    pass


# a condition trigger descriptor (base)
class _CondTriggerDescriptor:
    def __init__(self, trigger_name: str, cond_id: str):
        self._trigger_name = trigger_name
        self._cond_id = cond_id

    @property
    def trigger_name(self):
        return self._trigger_name

    @property
    def cond_id(self):
        return self._cond_id


# precondition trigger descriptor
class _PreCondTriggerDescriptor(_CondTriggerDescriptor):
    @property
    def type_str(self):
        return "pre"


# postcondition trigger descriptor
class _PostCondTriggerDescriptor(_CondTriggerDescriptor):
    @property
    def type_str(self):
        return "post"


# test method template for `LibPrePostCondsTestCase`
def _test(self: unittest.TestCase, descriptor: _CondTriggerDescriptor):
    # Execute:
    #
    #     $ conds-triggers run <trigger-name>
    #
    # where `<trigger-name>` is the descriptor trigger name.
    print("# Running trigger `{}`".format(descriptor.trigger_name))
    with subprocess.Popen(
        [_CONDS_TRIGGERS_PATH, "run", descriptor.trigger_name],
        stderr=subprocess.PIPE,
        universal_newlines=True,
    ) as proc:
        # wait for termination and get standard output/error data
        timeout = 5

        try:
            # wait for program end and get standard error pipe's contents
            _, stderr = proc.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            self.fail("Process hanged for {} seconds".format(timeout))
            return

        # assert that program aborted (only available on POSIX)
        if os.name == "posix":
            self.assertEqual(proc.returncode, -int(signal.SIGABRT))

        # assert that the standard error text contains the condition ID
        text = "Condition ID: `{}`.".format(descriptor.cond_id)
        self.assertIn(text, stderr)


# Condition trigger descriptors from the JSON array returned by
#
#     $ conds-triggers list
def _cond_trigger_descriptors_from_json(json_descr_array: tjson.ArrayVal):
    descriptors = []  # type: list[_CondTriggerDescriptor]
    descriptor_names = set()  # type: set[str]

    for json_descr in json_descr_array.iter(tjson.ObjVal):
        # sanity check: check for duplicate
        trigger_name = json_descr.at("name", tjson.StrVal).val

        if trigger_name in descriptor_names:
            raise ValueError(
                "Duplicate condition trigger name `{}`".format(trigger_name)
            )

        # condition ID
        cond_id = json_descr.at("cond-id", tjson.StrVal).val

        if cond_id.startswith("pre"):
            cond_type = _PreCondTriggerDescriptor
        elif cond_id.startswith("post"):
            cond_type = _PostCondTriggerDescriptor
        else:
            raise ValueError("Invalid condition ID `{}`".format(cond_id))

        descriptors.append(cond_type(trigger_name, cond_id))
        descriptor_names.add(trigger_name)

    return descriptors


# creates the individual tests of `LibPrePostCondsTestCase`
def _create_tests():
    # Execute `conds-triggers list` to get a JSON array of condition
    # trigger descriptors.
    json_descr_array = tjson.loads(
        subprocess.check_output(
            [_CONDS_TRIGGERS_PATH, "list"], universal_newlines=True
        ),
        tjson.ArrayVal,
    )

    # get condition trigger descriptor objects from JSON
    descriptors = _cond_trigger_descriptors_from_json(json_descr_array)

    # create test methods
    for descriptor in descriptors:
        # test method name
        test_meth_name = "test_{}".format(
            re.sub(r"[^a-zA-Z0-9_]", "_", descriptor.trigger_name)
        )

        # test method
        meth = functools.partialmethod(_test, descriptor)
        setattr(LibPrePostCondsTestCase, test_meth_name, meth)


_create_tests()


if __name__ == "__main__":
    unittest.main()
